import os
import sys
import json
from subprocess import call
import subprocess

from pip_box.dependencies_base import DependenciesBase

"""
    TODO
    - further testing is needed
"""

class PipUpgrade(DependenciesBase):
    def __init__(self, args):
        super(PipUpgrade, self).__init__()
        self.args = args        

        # Exclude editable and user defined packages
        self.excluded_pkgs = [] if self.args.exclude is None else self.args.exclude        
        if not self.args.local:     # Exclude editable packages
            self.excluded_pkgs = self.get_packages(args=['--editable']) + self.excluded_pkgs
        
        self.outdated = self.check_outdated()

    # Packages info
    
    def get_packages(self, args=[]):
        """
            This gets packages from pip, but it might be slower. Maybe use this later.
        """
        arg_list = [sys.executable, '-m', 'pip', 'list', '--format=json'] + args  

        packages = subprocess.check_output(arg_list)
        packages = packages.decode("utf-8").replace("\n", "")
        packages = json.loads(packages)

        packages_list = []

        for pkg in packages:
            packages_list.append(pkg['name'])

        return packages_list  

    def check_outdated(self):
        print('Checking outdated packages...')
        reqs = subprocess.check_output([sys.executable, '-m', 'pip', 'list', '--format=json', '--outdated'])
        reqs = reqs.decode("utf-8").replace("\n", "").replace("\r", "")     # fix here
        
        outdated = json.loads(reqs)     # List

        # Exclude package itself
        for i, item in enumerate(outdated):
            if item['name'] == 'pip-box':
                self.self_check = True
                outdated.pop(i)

        # Exclude other packages
        # print(f'Excluding locally installed packages: {self.excluded_pkgs}')
        outdated_return = []
        for i, item in enumerate(outdated):
            if not item['name'] in self.excluded_pkgs:
                outdated_return.append(item)

        return outdated_return    

    # Upgrade

    def clear_list(self, main, subtract, check_input_error=False):
        """
            Removes subtract's elements from main
        """
        for item in subtract:
            if item in main:
                main.pop(item)
            else:
                if check_input_error:
                    raise Exception(f'{item} is not in upgradable packages. This error is for safety incase of typos')
        return main
    
    def upgrade(self, be_upgraded):
        packages = []
        packages = {}
        for name, value in be_upgraded.items():
            if not value:
                packages[name] = ''
                pkg = name
            else:
                packages[name] = value[0][0] + value[0][1]
                pkg = name + value[0][0] + value[0][1]

        packages = self.clear_list(packages, self.wont_upgrade)

        if len(packages) > 0:
            # User input
            print(f'These packages will be upgraded: {list(packages.keys())}')
            cont_upgrade = input('Continue? (y/n): ')
            if cont_upgrade.lower() == 'y':
                cont_upgrade = True
            elif cont_upgrade.lower() == 'n':
                cont_upgrade = False
            elif cont_upgrade.startswith('-e'): # TODO take exclude arg here, test this further
                exclude = cont_upgrade.split(" ")
                exclude.remove('-e')
                self.clear_list(packages, exclude, check_input_error=True)
                cont_upgrade = True if len(packages) > 0 else False
            else:
                print('Please use one of the accepted inputs (y/n or -e PackageNames)\nCanceling...')
                cont_upgrade = False
            
            # Prepare packages dict
            packages = list(packages.items())
            packages = [''.join(x) for x in packages]
            
            if cont_upgrade:
                subprocess.check_call([sys.executable, '-m', 'pip', 'install', '-U', *packages])

        print('All packages are up to date! 🎉')
        
        if self.self_check:
            print("A new update avaliable for pip-box.\nPlease manually upgrade the tool using 'python -m pip install -U pip-box'")
            


import sys
import argparse

from pip_box.tool import PipUpgrade

parser = argparse.ArgumentParser()
parser.add_argument('-e', '--exclude', nargs='+', help="Exclude packages you don't want to upgrade")
parser.add_argument('--local', action='store_true', help="Upgrades local packages as well")
parser.add_argument('--novenv', action='store_true', help="Disables venv check")
parser.add_argument('-q', '--query', help="Query package dependency info from pypi")

args = parser.parse_args()


def check_venv():
    """
        Checks if virtualenv is active, throws an asssertion error if not
    """
    if not args.novenv:
        assert not sys.prefix == sys.base_prefix, 'Please use pip-box in a virtualenv. If you would like to surpass this use pip-box --novenv'

def main():
    check_venv()

    pip_upgrade = PipUpgrade(args)

    be_upgraded = pip_upgrade.get_dependencies()
    
    pip_upgrade.upgrade(be_upgraded)

if __name__ == "__main__":
    main()



from packaging import version


def version_check(apply_dep, latest_version, sign):
    # TODO find a better way
    # TODO check if it is correct

    sign_dict = {'==': 0, '~=': 1, '<': 2, '<=': 3, '>': 4, '>=': 5, '!=': 6}

    key = sign_dict[sign]

    if key == 0:
        result = version.parse(apply_dep) == version.parse(latest_version)
    elif key == 1:
        result = version.parse(apply_dep).minor == version.parse(latest_version).minor
    elif key == 2:
        result = version.parse(latest_version) < version.parse(apply_dep)
    elif key == 3:
        result = version.parse(latest_version) <= version.parse(apply_dep)
    elif key == 4:
        result = version.parse(latest_version) > version.parse(apply_dep)
    elif key == 5:
        result = version.parse(latest_version) >= version.parse(apply_dep)
    elif key == 6:
        result = version.parse(latest_version) != version.parse(apply_dep)
    else:
        print(f'version check error: {sign}')

    return result
    
def min_dependency(deps):
    """
        Gets min of two version
    """
    deps_list = []

    for dep in deps:
        _sign = dep[0]
        _version = dep[1]
        deps_list.append(version.parse(_version))

    return [_sign, str(min(deps_list))]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pip-box",
    version="0.4.4",
    author="Onur Cetinkol",
    author_email="realiti44@gmail.com",
    description="An easy tool for upgrading all of your packages while not breaking dependencies",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/realiti4/pip-box",
    entry_points = {
        'console_scripts': ['pip-box = pip_box:main',
                            'pip-upgrade = pip_box:main'],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    packages=["pip_box"],
    install_requires=["packaging"],
)



from pip._vendor import pkg_resources
from pip_box.version_checker import version_check, min_dependency

"""
    TODO 
    - Create a dependency matrix which will make searching much more easy
    - Update dep values by getting new info from pypi servers
"""


class DependenciesBase:
    def __init__(self):
        self.self_check = False

        self.packages = [dist.project_name for dist in pkg_resources.working_set]
        # self.packages.remove('pip')

        # self.packages = self.get_packages()
        # self.packages.remove('pip')
        
        self.dict = self.create_dict(self.packages)
        self.outdated = None

        self.importance_list = ['==', '~=', '<', '<=', '>', '>=', '!=']    
    
    def build_dep_matrix(self):
        """
            - One hot encoding for every package and use numpy or use pandas' indexes if we can search by name
        """
        raise NotImplementedError 

    def get_dependencies(self):
        """
            The main func for getting dependencies and comparing them to output a final list
        """

        self.retrieve_dependencies()

        be_upgraded = {}
        self.wont_upgrade = {}

        for pkg_dict in self.outdated:
            pkg_name = pkg_dict['name']
            current_version = pkg_dict['version']
            latest_version = pkg_dict['latest_version']

            deps = self.dict[pkg_name]

            apply_dep = self.compare_deps(pkg_name, deps, latest_version)

            be_upgraded[pkg_name] = apply_dep

            # TODO Check if it can be upgraded
            if len(apply_dep) > 0:
                version_ = apply_dep[0][1]
                sign_ = apply_dep[0][0]
                if not version_check(apply_dep[0][1], latest_version, apply_dep[0][0]):
                    self.wont_upgrade[pkg_name] = sign_ + version_

        return be_upgraded

    def compare_deps(self, pkg_name, deps, latest_version):
        """
            Compares dependencies in a list and decides what packages' final version should be
        """
        store = []
        done = False
        
        for i in self.importance_list:
            for index, i_dep in enumerate(deps):
                sign, version = i_dep

                # for i in self.importance_list:
                if sign == i:
                    store.append([sign, version])
                    done = True
                    # return store

            if done:
                if len(store) > 1:
                    # TODO check if taking min always is the right thing here
                    return [min_dependency(store)]
                else:
                    return store
        return store

    def retrieve_dependencies(self):
        """
            Retrieves dependencies pkg_main requires, and puts all dependent packages in self.dict with their version.
        """
        for pkg_main in self.packages:

            try:
                dep_list = pkg_resources.working_set.by_key[pkg_main].requires()
            except:
                dep_list = pkg_resources.working_set.by_key[pkg_main.lower()].requires()

            for i in dep_list:
                name = i.name        # Name of dependency
                specs = i.specs      # Specs of dependency
                
                if len(specs) != 0:
                    try:
                        if len(self.dict[name]) > 0:
                            self.dict[name].append(specs[0])
                        else:
                            self.dict[name] = specs
                    except:
                        if name == 'pillow':    # TODO fix lower case completely
                            self.dict['Pillow'] = specs
                        else:
                            print(f'Skipping {name}, warning: Name mismatch. This will be improved. Manually upgrade if needed')

    def create_dict(self, packages):
        return {x: [] for x in packages}
